import asyncio
import os
import signal
import sys

import httpx
from rich import print
from rich.progress import Progress, BarColumn
from rich.table import Table
from rich.panel import Panel
from rich.live import Live
from rich.traceback import install

from chouti.api.links import get_link
from chouti.config import config
from chouti.util.device import new_app_device
from chouti.util.task import Status, Task
from chouti.util.tracker import ClientVersionOutdatedError, Tracker
from chouti.util.archivist import get_archivist, new_archivist

install()

VERSION = "1.2"
BASE_CONCURRENCY = int(os.getenv("BASE_CONCURRENCY", 10))

interrupted = False

progress_table = Table.grid()
claimer_progress = Progress(
    "{task.description}",
    " * ",
    "{task.completed}/{task.total}",
)
processer_progress = Progress(
    "{task.description}",
    BarColumn(),
    "{task.completed}/{task.total}",
    "{task.elapsed:.2f}",
)

progress_table.add_row(
    Panel.fit(claimer_progress, title="Claimer Progress", border_style="green", padding=(2, 2)),
    Panel.fit(processer_progress, title="Processer Progress", border_style="blue", padding=(1, 2)),
)

done_tasks = 0

async def worker_claimer(name:int|str, tracker: Tracker, task_queue: asyncio.Queue[Task]):
    async def _remove_task(rich_task, after=0.3):
        await asyncio.sleep(after)
        claimer_progress.remove_task(rich_task)
    while True:
        if tracker.project.status.paused:
            claimer_progress.print("[tracker->client] Project paused, sleep 120s")
            await asyncio.sleep(120)
            continue
        if os.path.exists("stop") or interrupted:
            claimer_progress.print(f"[STOP]->Claimer({name}): stop file exists or interrupted, graceful exit...")
            return

        rich_task = claimer_progress.add_task(f"Claimer({name})", total=1)
        try:
            TASK = await tracker.claim_task_async()
        except ClientVersionOutdatedError as e:
            print(e)
            await interrupt(noforce=True) # stop
            continue
        except Exception as e:
            # traceback.print_exc()
            claimer_progress.update(rich_task, description=f"Claimer({name}): Failed to claim task: {e}", refresh=True)
            await asyncio.sleep(3)
            asyncio.create_task(_remove_task(rich_task))
            continue
        if TASK is None:
            notask_sleep = max(tracker.project.client.claim_task_delay * 10, 10)
            claimer_progress.update(rich_task, description=f"Claimer({name}): No tasks available, sleep", total=notask_sleep)
            # await asyncio.sleep(notask_sleep)
            for _ in range(int(notask_sleep)):
                if interrupted:
                    break
                await asyncio.sleep(1)
                claimer_progress.advance(rich_task)
            await asyncio.sleep(notask_sleep % 1)

            claimer_progress.remove_task(rich_task)
            continue

        # print(f"Claimed task {TASK.id}")
        claimer_progress.update(rich_task, description=f"Claimer({name}): Claimed task {TASK.id}", refresh=True, completed=1)
        await task_queue.put(TASK)

        # claimer_progress.remove_task(rich_task)

        asyncio.create_task(_remove_task(rich_task))

async def worker_process_task(name:int|str, tracker: Tracker, task_queue: asyncio.Queue[Task]):
    global done_tasks
    client = config.client
    while True:
        wait_task = processer_progress.add_task(f"Processors({name})<-get queue", total=1)
        TASK = await task_queue.get()
        processer_progress.remove_task(wait_task)
        rich_task = processer_progress.add_task(f"Processors({name})->Task({TASK.id})", total=4)
        processer_progress.advance(rich_task)

        assert isinstance(TASK.id, int)
        try:
            r = await get_link(client, TASK.id)
            processer_progress.advance(rich_task)
            r.raise_for_status()
            r_json = r.json()

            # insert item
            if "code" in r_json:
                if r_json["code"] == '400' and r_json["info"] == '找不到该帖子':
                    await tracker.insert_item_async(item_id=TASK.id, item_status="400", payload=None)
                else:
                    raise ValueError(f"Unknown code: {r_json}")
            elif "link" in r_json:
                # OK
                await tracker.insert_item_async(item_id=TASK.id, item_status=None, payload=r_json)
            else:
                raise ValueError(f"Unknown response: {r_json}")

            processer_progress.advance(rich_task)

            # update task status
            await tracker.update_task_async(task_id=TASK.id, status=Status.DONE)
            processer_progress.advance(rich_task)
        except Exception as e:
            # print(f"Task {TASK} failed: {e}")
            # traceback.print_exc()
            print(f"Task {TASK.id} failed: {e}")
            processer_progress.update(rich_task, description=f"Task {TASK.id} failed. (retrying)", refresh=True)
            
            # 抛的异常都会把任务设成 FAIL
            success = False
            for step in range(3, 60, 10):
                try:
                    processer_progress.advance(rich_task)
                    await tracker.update_task_async(task_id=TASK.id, status=Status.FAIL)
                    success = True
                    break
                except ClientVersionOutdatedError as e:
                    await interrupt(noforce=True)
                    print(e)
                    break
                except Exception as e:
                    # traceback.print_exc()
                    print(f"Task {TASK.id} failed to update status to FAIL, retry in {step}s")
                    await asyncio.sleep(step)
            if not success:
                # 重试失败，放弃，体面地退出 :)
                print(f"Task {TASK.id} failed to update status to FAIL, give up")
                print(f"Processer({name}) send interrupt signal")
                await interrupt(noforce=True)

        task_queue.task_done()
        done_tasks += 1
        
        async def _remove_task(rich_task):
            await asyncio.sleep(0.3)
            processer_progress.remove_task(rich_task)
        asyncio.create_task(_remove_task(rich_task))


async def show_status_progress(tracker: Tracker, task_queue: asyncio.Queue[Task]):
    queue_task = claimer_progress.add_task("Done/Queue", total=task_queue.qsize())
    while True:
        claimer_progress.update(queue_task,description=f"Done/Queue (QoS:{1 / tracker.project.client.claim_task_delay:.2f} T/s)", total=task_queue.qsize(), completed=done_tasks)
        await asyncio.sleep(0.1)

async def interrupt(noforce=False):
    """
    noforce: if True, will not force stop

    examples:
        1.
            interrupt() -> graceful exit
            interrupt() -> second interrupt, force exit
        2.
            interrupt() -> graceful exit
            interrupt(noforce=True) -> graceful exit
            interrupt() -> second interrupt, force exit
    """
    global interrupted
    if interrupted:
        if noforce:
            return
        print("[STOP] force stop")
        sys.exit(1)
    
    print("[STOP] interrupt signal received, graceful exit... (press again to force stop)")
    interrupted = True

async def interrupt_handler():
    global interrupted
    # handle keyboard interrupt, set `interrupted` to True
    try:
        loop = asyncio.get_event_loop()
        for signame in ('SIGINT', 'SIGTERM'):
            loop.add_signal_handler(getattr(signal, signame), lambda: asyncio.create_task(interrupt()))
    except NotImplementedError:
        pass


async def _main():
    headers = { "User-Agent": new_app_device() }
    config.client = httpx.AsyncClient(http2=True, http1=True, headers=headers, timeout=120)
    archivist = get_archivist() or new_archivist()

    # avoid infinite restart consume too much CPU
    await asyncio.sleep(1)

    tracker = Tracker(project_id="chouti_links", client_version=VERSION, archivist=archivist)

    task_queue: asyncio.Queue[Task] = asyncio.Queue(maxsize=BASE_CONCURRENCY * 2)

    await interrupt_handler()
    claimers = [worker_claimer(i, tracker, task_queue) for i in range(BASE_CONCURRENCY)]
    processors = [worker_process_task(i, tracker, task_queue) for i in range(BASE_CONCURRENCY)]

    with Live(progress_table, refresh_per_second=5):
        # start processors
        processors = [asyncio.create_task(p) for p in processors]
        # start claimers
        claimers = [asyncio.create_task(c) for c in claimers]
        # show queue progress
        status_progress = asyncio.create_task(show_status_progress(tracker, task_queue))
        # join claimers
        fus = await asyncio.gather(*claimers, return_exceptions=True)
        # join processors
        await task_queue.join()
        print("[STOP] All tasks in queue done.")
        for p in processors:
            p.cancel()

        print("[STOP] Stopped.")
        for fu in fus:
            if isinstance(fu, Exception):
                raise fu

def main():
    asyncio.run(_main())

if __name__ == "__main__":
    main()