import asyncio
import copy
from dataclasses import dataclass
import logging
import re
import json
import time
from typing import Any, Dict, Optional, Union

import httpx

from .task import Task

logger = logging.getLogger(__name__)

@dataclass
class ProjectMeta:
    identifier: str
    slug: str
    icon: str
    deadline: str
@dataclass
class ProjectStatus:
    public: bool
    paused: bool
@dataclass
class ProjectClient:
    version: str
    claim_task_delay: float
    """ 客户端在 claim_task 之后，多久之后才能再次 claim_task （软限制）"""
@dataclass
class ProjectMongodb:
    db_name: str
    item_collection: str
    queue_collection: str
    custom_doc_id_name: str
    """ TODO: 未来弃用 """
@dataclass
class Project:
    meta: ProjectMeta
    status: ProjectStatus
    client: ProjectClient
    mongodb: ProjectMongodb

    def __init__(self, meta: dict, status: dict, client: dict, mongodb: dict):
        self.meta = ProjectMeta(**meta)
        self.status = ProjectStatus(**status)
        self.client = ProjectClient(**client)
        self.mongodb = ProjectMongodb(**mongodb)


TRACKER_NODES = [
    "https://0.tracker.saveweb.org/",
    "https://1.tracker.saveweb.org/",
    "https://2.tracker.saveweb.org/",
    "https://3.tracker.saveweb.org/",
]

TEST_TRACKER_NODES = [
    "http://localhost:8080/",
]

class ClientVersionOutdatedError(Exception):
    pass

class Tracker:
    API_BASE: str = TEST_TRACKER_NODES[0]
    API_VERSION = "v1"
    client_version: str
    project_id: str
    """ [^a-zA-Z0-9_\\-] """
    archivist: str
    """ [^a-zA-Z0-9_\\-] """
    sync_session: httpx.Client
    session: httpx.AsyncClient
    project_id: str

    _claim_wait_lock = asyncio.Lock()

    __project: Project|None = None
    __project_last_fetched: float = 0

    __last_task_claimed_at: float = 0

    def _is_safe_string(self, s: str):
        r = re.compile(r"[^a-zA-Z0-9_\\-]")
        return not r.search(s)

    def __init__(self, project_id: str, client_version: str, archivist: str, session: httpx.AsyncClient|None = None):
        """
        raise: ClientVersionOutdatedError
        """
        assert self._is_safe_string(project_id) and self._is_safe_string(archivist), "[^a-zA-Z0-9_\\-]"
        self.project_id = project_id
        self.client_version = client_version
        self.archivist = archivist

        sync_transport = httpx.HTTPTransport(retries=3, http2=True, http1=True)
        transport = httpx.AsyncHTTPTransport(retries=3, http1=True, http2=True)

        self.sync_session = httpx.Client(http2=True, http1=True, timeout=120, transport=sync_transport)
        self.session = httpx.AsyncClient(http2=True, http1=True, timeout=120, transport=transport) if session is None else session

        self.select_best_tracker()

        if self.project.client.version != self.client_version:
            raise ClientVersionOutdatedError(f"client_version mismatch, please upgrade your client to {self.project.client.version}")

        print(f"[tracker] Hello, {self.archivist}!")
        print(f"[tracker] Project: {self.project}")
    
    @property
    def project(self):
        """
        return: Project.
        return __project cache if it's not expired (60s)
        """
        if self.__project is None or time.time() - self.__project_last_fetched > 60:
            try:
                self.__project = self.fetch_project(timeout=10)
            except Exception as e:
                print(f"[client->tracker] fetch_project failed. {e}")
                if self.__project is None:
                    raise e
            self.__project_last_fetched = time.time()
        assert self.__project is not None
        return self.__project
    
    def select_best_tracker(self):
        result = [] # [(node, latency_score_float)]
        print("[client->trackers] select_best_tracker...")
        # for node in TRACKER_NODES:
        #     try:
        #         self.sync_session.get(node + 'ping', timeout=0.05) # DNS preload, dirty hack
        #     except Exception:
        #         pass

        for node in TRACKER_NODES:
            start = time.time()
            print(f'--- {node} ---')
            try:
                print("[client->tracker]         \tping tracker...")
                r = self.sync_session.get(node + 'ping', timeout=5)
                r.raise_for_status()
                ping_time = time.time() - start
                print(f"[client<-tracker]         \tping tracker OK. {ping_time:.2f}s")

                print("[client->tracker->MongoDB]\tping MongoDB...")
                r = self.sync_session.get(node + 'ping_mongodb', timeout=8)
                print(r.text) # {"NearestEalapsedTime":236,"PrimaryEalapsedTime":298,"message":"MongoDB is up"}.
                r.raise_for_status() # 500 if MongoDB is down
                PrimaryEalapsedTime: int = r.json()["PrimaryEalapsedTime"]
                PrimaryEalapsedTimeSec = PrimaryEalapsedTime / 1000
                print(f"[client<-tracker<-MongoDB]\tping MongoDB OK. real latency: {ping_time+PrimaryEalapsedTimeSec:.2f}s")
                result.append((node, (ping_time*3)+(PrimaryEalapsedTimeSec*1))) # DB latency is not very impactful
            except Exception as e:
                print(f"[client->tracker          \tping failed. {e}")
                result.append((node, float('inf')))
        result.sort(key=lambda x: x[1])
        self.API_BASE = result[0][0]
        print("===============================")
        print(f"tracker selected: {self.API_BASE}")
        print("===============================")

    def fetch_project(self, timeout: int = 10):
        """
        return: Project, deep copy
        also refresh __project cache
        """
        logger.info(f"[client->tracker] fetch_project... {self.project_id}")
        r = self.sync_session.post(self.API_BASE + self.API_VERSION + '/project/' + self.project_id, timeout=timeout)
        r.raise_for_status()
        proj = Project(**r.json())
        self.__project = copy.deepcopy(proj)
        self.__project_last_fetched = time.time()
        logger.info(f"[client<-tracker] fetch_project. {self.project_id}")
        return proj

    def get_projects(self):
        logger.info(f"[client->tracker] get_projects.. {self.project_id}")
        r = self.sync_session.post(self.API_BASE + self.API_VERSION + '/projects')
        r.raise_for_status()
        # print("[client<-tracker] get_projects. OK")
        logger.info(f"[client<-tracker] get_projects. {self.project_id}")
        return [Project(**p) for p in r.json()]


    def _before_claim_task(self, with_delay: bool = True):
        if with_delay:
            if sleep_need := self.project.client.claim_task_delay - (time.time() - self.__last_task_claimed_at):
                if sleep_need > 0:
                    # print(f"[tracker] slow down you {sleep_need:.2f}s, Qos: {self.project.client.claim_task_delay}")
                    logger.info(f"[tracker] slow down you {sleep_need:.2f}s, Qos: {self.project.client.claim_task_delay}")
                    time.sleep(sleep_need)
                elif sleep_need < 0:
                    # print(f"[tracker] you are {sleep_need:.2f}s late, Qos: {self.project.client.claim_task_delay}")
                    logger.info(f"[tracker] you are {sleep_need:.2f}s late, Qos: {self.project.client.claim_task_delay}")

    async def _before_claim_task_async(self, with_delay: bool = True):
        if with_delay:
            if sleep_need := self.project.client.claim_task_delay - (time.time() - self.__last_task_claimed_at):
                if sleep_need > 0:
                    # print(f"[tracker] slow down you {sleep_need:.2f}s, Qos: {self.project.client.claim_task_delay}")
                    logger.info(f"[tracker] slow down you {sleep_need:.2f}s, Qos: {self.project.client.claim_task_delay}")
                    await self._claim_wait_lock.acquire()
                    await asyncio.sleep(sleep_need)
                    self._claim_wait_lock.release()
                elif sleep_need < 0:
                    logger.info(f"[tracker] you are {sleep_need:.2f}s late, Qos: {self.project.client.claim_task_delay}")

    def claim_task(self, with_delay: bool = True):
        self._before_claim_task(with_delay)

        logger.info("[client->tracker] claim_task")
        start = time.time()
        r = self.sync_session.post(f'{self.API_BASE}{self.API_VERSION}/project/{self.project_id}/{self.client_version}/{self.archivist}/claim_task')
        return self._after_claim_task(r, start)

    async def claim_task_async(self, with_delay: bool = True):
        """
        raise: ClientVersionOutdatedError
        """
        await self._before_claim_task_async(with_delay)

        # print("[client->tracker] claim_task")
        logger.info("[client->tracker] claim_task")
        start = time.time()
        r = await self.session.post(f'{self.API_BASE}{self.API_VERSION}/project/{self.project_id}/{self.client_version}/{self.archivist}/claim_task')
        return self._after_claim_task(r, start)

    def _after_claim_task(self, r: httpx.Response, start: float):
        """
        raise: ClientVersionOutdatedError
        """
        self.__last_task_claimed_at = time.time()
        if r.status_code == 404:
            # print(f'[client<-tracker] claim_task. (time cost: {time.time() - start:.2f}s):', r.text)
            logger.info(f'[client<-tracker] claim_task. (time cost: {time.time() - start:.2f}s):', r.text)
            return None # No tasks available
        if r.status_code == 200:
            r_json = r.json()
            # print(f'[client<-tracker] claim_task. OK (time cost: {time.time() - start:.2f}s):', r_json)
            logger.debug(f'[client<-tracker] claim_task. OK (time cost: {time.time() - start:.2f}s):', r_json)
            task = Task(**r_json)
            return task
        if r.status_code == 400:
            r_json = r.json()
            if 'error' in r_json and r_json['error'] == 'Client version not supported':
                raise ClientVersionOutdatedError(r.text)
        raise Exception(r.status_code, r.text)
    
    def _before_update_task(self, task_id: Union[str, int], status: str):
        assert isinstance(task_id, (str, int))
        assert isinstance(status, str)
        # print(f"[client->tracker] update_task task({task_id}) to status({status})")
        logger.info(f"[client->tracker] update_task task({task_id}) to status({status})")

    def update_task(self, task_id: Union[str, int], status: str):
        """
        task_id: 必须明确传入的是 int 还是 str
        status: 任务状态
        """
        self._before_update_task(task_id, status)

        r = self.sync_session.post(
            f'{self.API_BASE}{self.API_VERSION}/project/{self.project_id}/{self.client_version}/{self.archivist}/update_task/{task_id}', 
            data={
                'status': status,
                'task_id_type': 'int' if isinstance(task_id, int) else 'str'
            })
        return self._after_update_task(r, task_id)
    
    async def update_task_async(self, task_id: Union[str, int], status: str):
        """
        task_id: 必须明确传入的是 int 还是 str
        status: 任务状态

        raise: ClientVersionOutdatedError
        """
        self._before_update_task(task_id, status)

        r = await self.session.post(
            f'{self.API_BASE}{self.API_VERSION}/project/{self.project_id}/{self.client_version}/{self.archivist}/update_task/{task_id}', 
            data={
                'status': status,
                'task_id_type': 'int' if isinstance(task_id, int) else 'str'
            })
        return self._after_update_task(r, task_id)

    def _after_update_task(self, r: httpx.Response, task_id: Union[str, int])->Dict[str, Any]:
        if r.status_code == 200:
            r_json = r.json()
            # print(f'[client<-tracker] update_task task({task_id}). OK:', r_json)
            logger.info(f'[client<-tracker] update_task task({task_id}). OK:', r_json)
            return r_json
        if r.status_code == 400:
            r_json = r.json()
            if 'error' in r_json and r_json['error'] == 'Client version not supported':
                raise ClientVersionOutdatedError(r.text)
        raise Exception(r.text)
    
    def _before_insert_item(self, item_id: Union[str, int], item_status: Optional[Union[str, int]], payload: Optional[dict]):
        # item_id_type
        if isinstance(item_id, int):
            item_id_type = 'int'
        elif isinstance(item_id, str):
            item_id_type = 'str'
        else:
            raise ValueError("item_id must be int or str")

        if item_status is None:
            status_type = "None"
        elif isinstance(item_status, int):
            status_type = "int"
        elif isinstance(item_status, str):
            status_type = "str"
        else:
            raise ValueError("status must be int, str or None")
        payload_json_str = json.dumps(payload, ensure_ascii=False, separators=(',', ':'))
        # print(f"[client->tracker] insert_item item({item_id}), len(payload)={len(payload_json_str)}")
        logger.info(f"[client->tracker] insert_item item({item_id}), len(payload)={len(payload_json_str)}")

        return item_id_type, status_type, payload_json_str

    def insert_item(self, item_id: Union[str, int], item_status: Optional[Union[str, int]] = None, payload: Optional[dict] = None):
        """
        item_id: 必须明确传入的是 int 还是 str
        item_status: item 状态，而不是 task 状态。可用于标记一些被删除、被隐藏的 item 之类的。就不用添加一堆错误响应到 payload 里了。 
        payload: 可以传入任意的可转为 ext-json 的对象 (包括 None)
        """
        item_id_type, status_type, payload_json_str = self._before_insert_item(item_id, item_status, payload)

        r = self.sync_session.post(
            f'{self.API_BASE}{self.API_VERSION}/project/{self.project_id}/{self.client_version}/{self.archivist}/insert_item/{item_id}',
            data={
                'item_id_type': item_id_type,
                'item_status': item_status,
                'item_status_type': status_type,
                'payload': payload_json_str
            })
        return self._after_insert_item(r, item_id)
    
    async def insert_item_async(self, item_id: Union[str, int], item_status: Optional[Union[str, int]] = None, payload: Optional[dict] = None):
        """
        item_id: 必须明确传入的是 int 还是 str
        item_status: item 状态，而不是 task 状态。可用于标记一些被删除、被隐藏的 item 之类的。就不用添加一堆错误响应到 payload 里了。 
        payload: 可以传入任意的可转为 ext-json 的对象 (包括 None)
        """
        item_id_type, status_type, payload_json_str = self._before_insert_item(item_id, item_status, payload)

        r = await self.session.post(
            f'{self.API_BASE}{self.API_VERSION}/project/{self.project_id}/{self.client_version}/{self.archivist}/insert_item/{item_id}',
            data={
                'item_id_type': item_id_type,
                'item_status': item_status,
                'item_status_type': status_type,
                'payload': payload_json_str
            })
        return self._after_insert_item(r, item_id)

    def _after_insert_item(self, r: httpx.Response, item_id: Union[str, int])->Dict[str, Any]:
        if r.status_code == 200:
            r_json = r.json()
            # print(f'[client<-tracker] insert_item item({item_id}). OK:', r_json)
            logger.info(f'[client<-tracker] insert_item item({item_id}). OK:', r_json)
            return r_json
        raise Exception(r.text)
